import asyncio
import logging

from .protocol import BasicProtocol, create_peer

logger = logging.getLogger(__name__)


def mainloop(receptor, ping_interval=None, loop=asyncio.get_event_loop(), skip_run=False):
    config = receptor.config
    if config.server.server_enable:
        listener = loop.create_server(
            lambda: BasicProtocol(receptor, loop),
            config.server.address, config.server.port, ssl=config.get_server_ssl_context())
        loop.create_task(listener)
        logger.info("Serving on %s:%s", config.server.address, config.server.port)
    for peer in config.peers:
        loop.create_task(create_peer(receptor, loop, *peer.split(":", 1)))
    if ping_interval:
        ping_time = (((int(loop.time()) + 1) // ping_interval) + 1) * ping_interval
        loop.call_at(ping_time, loop.create_task, send_pings_and_reschedule(receptor, loop, ping_time, ping_interval))
    if not skip_run:
        try:
            loop.run_until_complete(receptor.shutdown_handler())
        except KeyboardInterrupt:
            pass
        finally:
            loop.stop()


async def send_pings_and_reschedule(receptor, loop, ping_time, ping_interval):
    logger.debug(f'Scheduling mesh ping.')
    for node_id in receptor.router.get_nodes():
        await receptor.router.ping_node(node_id)
    loop.call_at(ping_time + ping_interval, 
                 loop.create_task, send_pings_and_reschedule(
                     receptor, loop, ping_time + ping_interval, ping_interval))



import argparse
import datetime
import logging
import logging.config
from .config import ReceptorConfig, DEFAULT_CONFIG
from . import node
from . import controller
from receptor import Receptor

logger = logging.getLogger(__name__)


def map_args_to_config(args):
    to_return = {}
    if getattr(args, 'listen_address', None):
        to_return.setdefault('server', {})['address'] = args.listen_address
    if getattr(args, 'listen_port', None):
        to_return.setdefault('server', {})['port'] = args.listen_port
    if getattr(args, 'server_enable', None):
        to_return.setdefault('server', {})['server_enable'] = args.server_enable
    if getattr(args, 'debug', None):
        to_return.setdefault('server', {})['debug'] = args.debug
    if getattr(args, 'ssl_certificate', None):
        to_return.setdefault('server', {})['ssl_certificate'] = args.ssl_certificate
    if getattr(args, 'ssl_key', None):
        to_return.setdefault('server', {})['ssl_key'] = args.ssl_key
    if getattr(args, 'peer', None):
        to_return['peers'] = {peer: '' for peer in args.peer}
    if getattr(args, 'node_id', None):
        to_return.setdefault('receptor', {})['node_id'] = args.node_id
    return to_return


def run_as_controller(args):
    config = ReceptorConfig(args.config, map_args_to_config(args))
    receptor = Receptor(config)
    logger.info(f'Starting up as node ID {receptor.node_id}')
    controller.mainloop(receptor, args.socket_path)


def run_as_ping(args):
    logger.info(f'Sending ping to {args.recipient}.')
    now = datetime.datetime.utcnow()
    controller.send_directive('receptor:ping', args.recipient, now.isoformat(), args.socket_path)


def run_as_send(args):
    logger.info(f'Sending a {args.directive} directive to {args.recipient}.')
    controller.send_directive(args.directive, args.recipient, args.payload, args.socket_path)


def run_as_node(args):
    config = ReceptorConfig(args.config, map_args_to_config(args))
    receptor = Receptor(config)
    logger.info("Running as Receptor node with ID: {}".format(receptor.node_id))
    node.mainloop(receptor, args.ping_interval)


def main(args=None):
    parser = argparse.ArgumentParser("receptor")
    parser.add_argument(
        "-c", "--config", default="/etc/receptor/receptor.conf",
        help='Path to configuration file')
    parser.add_argument(
        "--debug", action="store_true", default=False,
        help='Emit debugging output')
    parser.add_argument(
        '--ssl_certificate',
        help="Certificate Chain File"
    )
    parser.add_argument(
        '--ssl_key',
        help="Certificate Key File"
    )
    subparsers = parser.add_subparsers(
        title='subcommands')
    
    subparser_node = subparsers.add_parser(
        'node',
        help='Run a Receptor node')
    subparser_node.add_argument(
        "--listen-address",
        help=f'Set/override IP address to listen on. If not set here or in a config file, the default is {DEFAULT_CONFIG["server"]["address"]}')
    subparser_node.add_argument(
        "--listen-port",
        help=f'Set/override TCP port to listen on. If not set here or in a config file, the default is {DEFAULT_CONFIG["server"]["port"]}')
    subparser_node.add_argument(
        "-p", "--peer", action='append',
        help=f'Set/override peer nodes/controllers to connect to. Use multiple times for multiple peers.')
    subparser_node.add_argument(
        "--node-id",
        help='Set/override node identifier. If unspecified here or in a config file, one will be automatically generated.')
    subparser_node.add_argument(
        "--server-disable", action="store_true", default=False,
        help="Disable the server function and only connect to configured peers.")
    subparser_node.add_argument(
        '--ping-interval', metavar="N", type=int,
        help="If specified, the node will ping all other known nodes in the mesh every N seconds."
    )
    subparser_node.set_defaults(func=run_as_node)

    subparser_controller = subparsers.add_parser(
        'controller',
        help='Run a Receptor controller'
    )
    subparser_controller.add_argument(
        '--socket-path', default='/var/run/receptor_controller.sock',
        help='Path to control socket'
    )
    subparser_controller.add_argument(
        "--listen-address",
        help=f'Set/override IP address to listen on. If not set here or in a config file, the default is {DEFAULT_CONFIG["server"]["address"]}')
    subparser_controller.add_argument(
        "--listen-port",
        help=f'Set/override TCP port to listen on. If not set here or in a config file, the default is {DEFAULT_CONFIG["server"]["port"]}')
    subparser_controller.add_argument(
        "--node-id",
        help='Set/override node identifier. If unspecified here or in a config file, one will be automatically generated.')
    subparser_controller.set_defaults(func=run_as_controller)

    subparser_ping = subparsers.add_parser(
        'ping',
        help='Tell the local controller to ping a node'
    )
    subparser_ping.add_argument(
        '--socket-path', default='/var/run/receptor_controller.sock',
        help='Path to control socket'
    )
    subparser_ping.add_argument(
        'recipient',
        help='Node ID of the Receptor node or controller to ping'
    )
    subparser_ping.set_defaults(func=run_as_ping)

    subparser_send = subparsers.add_parser(
        'send',
        help='Send a directive to a node'
    )
    subparser_send.add_argument(
        '--socket-path', default='/var/run/receptor_controller.sock',
        help='Path to control socket'
    )
    subparser_send.add_argument(
        '--directive',
        help='Directive to send'
    )
    subparser_send.add_argument(
        '--recipient',
        help='Node ID of the Receptor node or controller to direct'
    )
    subparser_send.add_argument(
        'payload',
        help='Payload of the directive to send. Use - for stdin.'
    )
    subparser_send.set_defaults(func=run_as_send)
   

    args = parser.parse_args(args)
    
    logging.config.dictConfig(
        {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'verbose': {
                    'format': '{levelname} {asctime} {module} {message}',
                    'style': '{',
                }
            },
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose'
                },
            },
            'loggers': {
                'receptor': {
                    'handlers': ['console'],
                    'level': 'DEBUG' if args.debug else 'INFO',
                },
            },
        }
    )
    args.func(args)


if __name__ == '__main__':
    # We were run with python -m
    main()



import base64
import datetime
import json
import logging
import uuid

logger = logging.getLogger(__name__)


class OuterEnvelope:
    def __init__(self, frame_id, sender, recipient, route_list, inner):
        self.frame_id = frame_id
        self.sender = sender
        self.recipient = recipient
        self.route_list = route_list
        self.inner = inner
        self.inner_obj = None

    async def deserialize_inner(self, receptor):
        self.inner_obj = await InnerEnvelope.deserialize(receptor, self.inner)

    @classmethod
    def from_raw(cls, raw):
        doc = json.loads(raw)
        return cls(**doc)
    
    def serialize(self):
        return json.dumps(dict(
            frame_id=self.frame_id,
            sender=self.sender,
            recipient=self.recipient,
            route_list=self.route_list,
            inner=self.inner
        ))


class InnerEnvelope:
    def __init__(self, receptor, message_id, sender, recipient, message_type, timestamp,
                 raw_payload, directive=None, in_response_to=None, ttl=None,
                 serial=1):
        self.receptor = receptor
        self.message_id = message_id
        self.sender = sender
        self.recipient = recipient
        self.message_type = message_type # 'directive' or 'response'
        self.timestamp = timestamp # ISO format
        self.raw_payload = raw_payload
        self.directive = directive # None if response, 'namespace:action' if not
        self.in_response_to = in_response_to # None if directive, a message_id if not
        self.ttl = ttl # Optional
        self.serial = serial # serial index of responses

    @classmethod
    async def deserialize(cls, receptor, msg):
        payload = await receptor.config.components.security_manager.verify_msg(msg)
        # validate msg
        # msg+sig
        return cls(receptor=receptor, **json.loads(payload))

    @classmethod
    def make_response(cls, receptor, recipient, payload, in_response_to, serial, ttl=None):
        if isinstance(payload, bytes):
            encoded_payload = base64.encodebytes(payload)
        else:
            encoded_payload = payload
        return cls(
            receptor=receptor,
            message_id=str(uuid.uuid4()),
            sender=receptor.node_id,
            recipient=recipient,
            message_type='response',
            timestamp=datetime.datetime.utcnow().isoformat(),
            raw_payload=encoded_payload,
            directive=None,
            in_response_to=in_response_to,
            ttl=ttl,
            serial=serial
        )

    def sign_and_serialize(self):
        return self.receptor.config.components.security_manager.sign_response(self)



import json
import logging
logger = logging.getLogger(__name__)


class MallCop:

    async def verify_node(self, node):
        return True

    async def verify_controller(self, controller):
        return True

    async def verify_msg(self, msg):
        return msg

    async def verify_directive(self, directive):
        return True

    async def verify_response(self, response):
        return True
    
    async def sign_response(self, inner_envelope):
        return json.dumps(
            {attr: getattr(inner_envelope, attr)
             for attr in ['message_id', 'sender', 'recipient', 'message_type',
                          'timestamp', 'raw_payload', 'directive',
                          'in_response_to', 'ttl', 'serial']}
        )




import collections
import logging

from .base import BaseBufferManager, BaseBuffer

logger = logging.getLogger(__name__)


class InMemoryBufferManager(BaseBufferManager):
    _buffers = {}

    def get_buffer_for_node(self, node_id):
        return self._buffers.setdefault(node_id, InMemoryBuffer(node_id))


# NOTE: This is not thread safe
class InMemoryBuffer(BaseBuffer):
    def __init__(self, node_id):
        super().__init__(node_id)
        self._buffer = collections.deque()

    def push(self, message):
        logger.debug(f'Pushing a message to {self.node_id} buffer.')
        self._buffer.append(message)
    
    def pop(self):
        return self._buffer.popleft()
    
    def flush(self):
        logger.debug(f'Flushing buffer for {self.node_id}')
        to_return = list(self._buffer)
        self._buffer.clear()
        return to_return



import importlib
import configparser
import ssl
import logging

logger = logging.getLogger(__name__)

DEFAULT_CONFIG = dict(
    server=dict(
        port=8888,
        address='0.0.0.0',
        server_enable=True,
        debug=False,
        ssl_certificate='',
        ssl_key='',
    ),
    receptor=dict(
        node_id=None,
    ),
    peers=dict(),
    components=dict(
        security_manager='receptor.security.MallCop',
        buffer_manager='receptor.buffers.memory.InMemoryBufferManager'
    ),
)
VALUELESS_SECTIONS = ['peers']
SINGLETONS = {}


def py_class(class_spec):
    if class_spec not in SINGLETONS:
        module_name, class_name = class_spec.rsplit('.', 1)
        module_obj = importlib.import_module(module_name)
        class_obj = getattr(module_obj, class_name)
        SINGLETONS[class_spec] = class_obj()
    return SINGLETONS[class_spec]


CAST_MAP = dict(
    server=dict(
        port=int,
        server_enable=lambda val: val == "True",
        debug=lambda val: val == "True",
    ),
    peers=dict(),
    receptor=dict(),
    components=dict(
        security_manager=py_class,
        buffer_manager=py_class
    )
)


class ReceptorConfigSection:
    def __init__(self, parser, section):
        self._parser = parser
        self._section = section
    
    def __getattr__(self, key):
        if not self._parser.has_section(self._section):
            return None
        to_return = self._parser.get(self._section, key)
        cast_fn = CAST_MAP.get(self._section, {}).get(key, None)
        if cast_fn:
            return cast_fn(to_return)
        return to_return


class ReceptorConfig:
    def __init__(self, config_path=None, cmdline_args=None):
        self._parser = configparser.ConfigParser(allow_no_value=True, delimiters=('=',))
        self._parser.read_dict(DEFAULT_CONFIG)
        if config_path:
            self._parser.read([config_path])
        if cmdline_args:
            self._parser.read_dict(cmdline_args)

    def get_client_ssl_context(self):
        if self.server.ssl_certificate:
            logger.debug("Loading SSL Client Context")
            return ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=self.server.ssl_certificate)
        else:
            return None

    def get_server_ssl_context(self):
        if self.server.ssl_certificate and self.server.ssl_key:
            sc = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            sc.load_cert_chain(self.server.ssl_certificate, self.server.ssl_key)
            logger.debug("Loading SSL Server Context")
            return sc
        else:
            return None
    
    def __getattr__(self, key):
        if key in VALUELESS_SECTIONS:
            return list(dict(self._parser.items(key)).keys())
        return ReceptorConfigSection(self._parser, key)



import datetime
import asyncio
import logging
import json
import uuid
from collections import deque
from .messages import envelope

logger = logging.getLogger(__name__)

DELIM = b"\x1b[K"
SIZEB = b"\x1b[%dD"


class DataBuffer:
    def __init__(self, deserializer=json.loads):
        self.q = deque()
        self._buf = b""
        self.deserializer = deserializer

    def add(self, data):
        self._buf = self._buf + data
        *ready, self._buf = self._buf.rsplit(DELIM)
        self.q.extend(ready)

    def get(self):
        while self.q:
            yield self.deserializer(self.q.popleft())


class BaseProtocol(asyncio.Protocol):
    def __init__(self, receptor, loop):
        self.receptor = receptor
        self.loop = loop

    async def watch_queue(self, node, transport):
        buffer_mgr = self.receptor.config.components.buffer_manager
        buffer_obj = buffer_mgr.get_buffer_for_node(node)
        while not transport.is_closing():
            try:
                msg = buffer_obj.pop()
                transport.write(msg + DELIM)
            except IndexError:
                await asyncio.sleep(1)
            except Exception as e:
                logger.exception("Error received trying to write to {}: {}".format(node, e))
                buffer_obj.push(msg)
                transport.close()
                return

    def connection_made(self, transport):
        self.peername = transport.get_extra_info('peername')
        self.transport = transport
        self.greeted = False
        self._buf = DataBuffer()
        self.loop.create_task(self.consume())

    def connection_lost(self, exc):
        self.receptor.remove_connection(self)

    def data_received(self, data):
        logger.debug(data)
        self._buf.add(data)

    async def consume(self):
        while not self.greeted:
            logger.debug('Looking for handshake...')
            for data in self._buf.get():
                logger.debug(data)
                if data["cmd"] == "HI":
                    self.handle_handshake(data)
                    break
                else:
                    logger.error("Handshake failed!")
            await asyncio.sleep(.1)
        logger.debug("handshake complete, starting normal handle loop")
        self.loop.create_task(self.connection.handle_loop(self._buf))

    def handle_handshake(self, data):
        self.greeted = True
        self.connection = self.receptor.add_connection(data["id"], self)
        self.loop.create_task(self.watch_queue(data["id"], self.transport))
        self.loop.create_task(self.connection.handle_loop(self._buf))

    def send_handshake(self):
        msg = json.dumps({
            "cmd": "HI",
            "id": self.receptor.node_id,
        }).encode("utf-8")
        self.transport.write(msg + DELIM)


class BasicProtocol(BaseProtocol):
    def connection_made(self, transport):
        super().connection_made(transport)
        logger.info('Connection from {}'.format(self.peername))

    def handle_handshake(self, data):
        super().handle_handshake(data)
        logger.debug("Received handshake from client with id %s, responding...", data["id"])
        self.send_handshake()
        self.connection.send_route_advertisement()


async def create_peer(receptor, loop, host, port):
    while True:
        try:
            await loop.create_connection(
                lambda: BasicClientProtocol(receptor, loop), host, port, ssl=receptor.config.get_client_ssl_context())
            break
        except Exception:
            logger.exception("Connection Refused: {}:{}".format(host, port))
            await asyncio.sleep(5)


class BasicClientProtocol(BaseProtocol):
    def connection_made(self, transport):
        super().connection_made(transport)
        logger.info("Connection to %s", self.peername)
        logger.debug("Sending handshake to server...")
        self.send_handshake()

    def connection_lost(self, exc):
        logger.info('Connection lost with the server...')
        super().connection_lost(exc)
        info = self.transport.get_extra_info('peername')
        self.loop.create_task(create_peer(self.receptor, self.loop, info[0], info[1]))

    def handle_handshake(self, data):
        super().handle_handshake(data)
        logger.debug("Received handshake from server with id %s", data["id"])
        self.connection.send_route_advertisement()


class BasicControllerProtocol(asyncio.Protocol):

    def __init__(self, receptor, loop):
        self.receptor = receptor
        self.loop = loop

    def connection_made(self, transport):
        self.transport = transport
        if self not in self.receptor.controller_connections:
            self.receptor.controller_connections.append(self)

    def connection_lost(self, exc):
        if self in self.receptor.controller_connections:
            self.receptor.controller_connections.remove(self)

    def emit_response(self, response):
        self.transport.write(json.dumps(dict(timestamp=response.timestamp,
                                             in_response_to=response.in_response_to,
                                             payload=response.raw_payload)).encode())

    def data_received(self, data):
        recipient, directive, payload = data.rstrip(DELIM).decode('utf8').split('\n', 2)
        message_id = str(uuid.uuid4())
        logger.info(f'{message_id}: Sending {directive} to {recipient}')
        sent_timestamp = datetime.datetime.utcnow()
        inner_env = envelope.InnerEnvelope(
            receptor=self.receptor,
            message_id=message_id,
            sender=self.receptor.node_id,
            recipient=recipient,
            message_type='directive',
            timestamp=sent_timestamp.isoformat(),
            raw_payload=payload,
            directive=directive
        )
        # TODO: Response expiration task?
        # TODO: Persistent registry?
        self.loop.create_task(self.receptor.router.send(inner_env,
                                                        expected_response=True))



import logging
logger = logging.getLogger(__name__)


class BaseBufferManager:
    def get_buffer_for_node(self, node_id):
        raise NotImplementedError()


class BaseBuffer:
    node_id = None

    def __init__(self, node_id):
        self.node_id = node_id

    def push(self, message):
        raise NotImplementedError()
    
    def pop(self):
        raise NotImplementedError()
    
    def flush(self):
        raise NotImplementedError()



import logging

import datetime
from collections import defaultdict
import heapq
import random
import uuid

from dateutil import parser
from .messages import envelope
from .exceptions import UnrouteableError

logger = logging.getLogger(__name__)


async def log_ping(response):
    pong_received = datetime.datetime.utcnow()
    ping_sent, ping_received = response.raw_payload.split('|')
    ping_time = parser.parse(ping_received) - parser.parse(ping_sent)
    pong_time = pong_received - parser.parse(ping_received)
    logger.info(f'Ping report for {response.sender}: '
                f'ping={ping_time.total_seconds()}s; '
                f'pong={pong_time.total_seconds()}s')


class MeshRouter:
    _nodes = set()
    _edges = set()
    response_registry = dict()

    def __init__(self, receptor):
        self.receptor = receptor
        self.node_id = receptor.node_id

    def debug_router(self):
        logger.debug("Receptor Edges: {}".format(self._edges))
        if self.receptor.config.server.debug:
            fd = open("graph_{}.dot".format(self.receptor.node_id), "w")
            fd.write("graph {")
            for left, right, weight in self._edges:
                fd.write("{} -- {};".format(left, right))
            fd.write("}")
            fd.close()

    def node_is_known(self, node_id):
        return node_id in self._nodes or node_id == self.node_id

    def find_edge(self, left, right):
        node_actual = sorted([left, right])
        for edge in self._edges:
            if node_actual[0] == edge[0] and node_actual[1] == edge[1]:
                return edge
        return None

    def register_edge(self, left, right, cost):
        if left != self.node_id:
            self._nodes.add(left)
        if right != self.node_id:
            self._nodes.add(right)
        edge = self.update_node(left, right, cost)
        if not edge:
            self._edges.add((*sorted([left, right]), cost))
        self.debug_router()

    def update_node(self, left, right, cost):
        edge = self.find_edge(left, right)
        if edge:
            new_edge = (edge[0], edge[1], cost)
            self._edges.remove(edge)
            self._edges.add(new_edge)
            return edge
        return None

    def remove_node(self, node):
        edge = self.find_edge(self.node_id, node)
        if edge:
            self._edges.remove(edge)
            return edge
        return None

    def get_edges(self):
        """Returns set of edges"""
        return list(self._edges)
    
    def get_nodes(self):
        return self._nodes
    
    async def ping_node(self, node_id, callback=log_ping):
        logger.info(f'Sending ping to node {node_id}')
        now = datetime.datetime.utcnow().isoformat()
        ping_envelope = envelope.InnerEnvelope(
            receptor=self.receptor,
            message_id=str(uuid.uuid4()),
            sender=self.node_id,
            recipient=node_id,
            message_type='directive',
            timestamp=now,
            raw_payload=now,
            directive='receptor:ping',
            ttl=15
        )
        await self.send(ping_envelope, callback)

    def find_shortest_path(self, to_node_id):
        """Implementation of Dijkstra algorithm"""
        cost_map = defaultdict(list)
        for left, right, cost in self._edges:
            cost_map[left].append((cost, right))
            cost_map[right].append((cost, left))

        heap, seen, mins = [(0, self.node_id, [])], set(), {self.node_id: 0}
        while heap:
            (cost, vertex, path) = heapq.heappop(heap)
            if vertex not in seen:
                seen.add(vertex)
                path = [vertex] + path                
                if vertex == to_node_id:
                    logger.debug(f'Shortest path to {to_node_id} with cost {cost} is {path}')
                    return path
                cost_map_for_vertex = cost_map.get(vertex, ())
                random.shuffle(cost_map_for_vertex)
                for next_cost, next_vertex in cost_map.get(vertex, ()):
                    if next_vertex in seen:
                        continue
                    min_so_far = mins.get(next_vertex, None)
                    next_total_cost = cost + next_cost
                    if min_so_far is None or next_total_cost < min_so_far:
                        mins[next_vertex] = next_total_cost
                        heapq.heappush(heap, (next_total_cost, next_vertex, path))

    async def forward(self, outer_envelope, next_hop):
        """
        Forward a message on to the next hop closer to its destination
        """
        buffer_mgr = self.receptor.config.components.buffer_manager
        buffer_obj = buffer_mgr.get_buffer_for_node(next_hop)
        outer_envelope.route_list.append(self.node_id)
        logger.debug(f'Forwarding frame {outer_envelope.frame_id} to {next_hop}')
        buffer_obj.push(outer_envelope.serialize().encode("utf-8"))


    def next_hop(self, recipient):
        """
        Return the node ID of the next hop for routing a message to the
        given recipient. If the current node is the recipient or there is
        no path, then return None.
        """
        if recipient == self.node_id:
            return None
        path = self.find_shortest_path(recipient)
        if path:
            return path[-2]


    async def send(self, inner_envelope, expected_response=False):
        """
        Send a new message with the given outer envelope.
        """
        next_node_id = self.next_hop(inner_envelope.recipient)
        if not next_node_id:
            raise UnrouteableError(f'No route found to {inner_envelope.recipient}')
        signed = await inner_envelope.sign_and_serialize()
        outer_envelope = envelope.OuterEnvelope(
            frame_id=str(uuid.uuid4()),
            sender=self.node_id,
            recipient=inner_envelope.recipient,
            route_list=[self.node_id],
            inner=signed
        )
        logger.debug(f'Sending {inner_envelope.message_id} to {inner_envelope.recipient} via {next_node_id}')
        if expected_response and inner_envelope.message_type == 'directive':
            self.response_registry[inner_envelope.message_id] = dict(message_sent_time=inner_envelope.timestamp)
        await self.forward(outer_envelope, next_node_id)






import datetime
import logging

from ..exceptions import UnknownDirective
from . import envelope

logger = logging.getLogger(__name__)


class Directive:
    def __init__(self, type_, payload):
        self.type_ = type_
        self.payload = payload


class Control:
    CONTROL_DIRECTIVES = ['ping']

    async def __call__(self, router, inner_env):
        _, action = inner_env.directive.split(':', 1)
        if action not in self.CONTROL_DIRECTIVES:
            raise UnknownDirective(f'Unknown control directive: {action}')
        action_method = getattr(self, action)
        responses = action_method(inner_env)
        serial = 0
        async for response in responses:
            serial += 1
            enveloped_response = envelope.InnerEnvelope.make_response(
                receptor=router.receptor,
                recipient=inner_env.sender,
                payload=response,
                in_response_to=inner_env.message_id,
                serial=serial
            )
            await router.send(enveloped_response)
    
    async def ping(self, inner_env):
        logger.info(f'Received ping from {inner_env.sender}')
        yield f'{inner_env.raw_payload}|{datetime.datetime.utcnow().isoformat()}'


control = Control()



import pkg_resources
import logging

from . import exceptions
from .messages import envelope
logger = logging.getLogger(__name__)


class WorkManager:
    def __init__(self, receptor):
        self.receptor = receptor

    def load_receptor_worker(self, name):
        entry_points = [x for x in filter(lambda x: x.name == name,
                                          pkg_resources.iter_entry_points("receptor.worker"))]
        if not entry_points:
            raise exceptions.UnknownDirective(f"Error loading directive handlers for {name}")
        return entry_points[0].load()

    async def handle(self, inner_env):
        logger.info(f'Handling work for {inner_env.message_id} as {inner_env.directive}')
        namespace, action = inner_env.directive.split(':', 1)
        worker_module = self.load_receptor_worker(namespace)
        try:
            action_method = getattr(worker_module, f'{action}')
        except AttributeError:
            logger.exception(f'Could not load action {action} from {namespace}')
            raise exceptions.InvalidDirectiveAction(f'Invalid action {action} for {namespace}')
        responses = action_method(inner_env)
        serial = 0
        async for response in responses:
            serial += 1
            logger.debug(f'Response emitted for {inner_env.message_id}, serial {serial}')
            enveloped_response = envelope.InnerEnvelope.make_response(
                receptor=self.receptor,
                recipient=inner_env.sender,
                payload=response,
                in_response_to=inner_env.message_id,
                serial=serial
            )
            await self.receptor.router.send(enveloped_response)





import os
import uuid
import asyncio
import logging
import json

from .config import ReceptorConfig
from .router import MeshRouter
from .work import WorkManager
from . import exceptions
from .messages import envelope, directive

logger = logging.getLogger(__name__)

RECEPTOR_DIRECTIVE_NAMESPACE = 'receptor'


class Connection:
    def __init__(self, id_, protocol_obj, buffer_mgr, receptor):
        self.id_ = id_
        self.protocol_obj = protocol_obj
        self.buffer_mgr = buffer_mgr
        self.receptor = receptor

    def __str__(self):
        return f"<Connection {self.id_} {self.protocol_obj}>"

    async def handle_loop(self, buf):
        while True:
            for data in buf.get():
                if "cmd" in data and data["cmd"] == "ROUTE":
                    self.handle_route_advertisement(data)
                else:
                    await self.handle_message(data)
            await asyncio.sleep(.1)

    def handle_route_advertisement(self, data):
        for edge in data["edges"]:
            existing_edge = self.receptor.router.find_edge(edge[0], edge[1])
            if existing_edge and existing_edge[2] > edge[2]:
                self.receptor.router.update_node(edge[0], edge[1], edge[2])
            else:
                self.receptor.router.register_edge(*edge)
        self.send_route_advertisement(data["edges"], data["seen"])

    def send_route_advertisement(self, edges=None, seen=[]):
        edges = edges or self.receptor.router.get_edges()
        seen = set(seen)
        logger.debug("Emitting Route Advertisements, excluding {}".format(seen))
        destinations = set(self.receptor.connections) - seen
        seens = list(seen | destinations | {self.receptor.node_id})

        # TODO: This should be a broadcast call to the connection manager
        for target in destinations:
            buf = self.buffer_mgr.get_buffer_for_node(target)
            buf.push(json.dumps({
                "cmd": "ROUTE",
                "id": self.receptor.node_id,
                "edges": edges,
                "seen": seens
            }).encode("utf-8"))

    async def handle_message(self, msg):
        outer_env = envelope.OuterEnvelope(**msg)
        next_hop = self.receptor.router.next_hop(outer_env.recipient)
        if next_hop is None:
            await outer_env.deserialize_inner(self.receptor)
            if outer_env.inner_obj.message_type == 'directive':
                namespace, _ = outer_env.inner_obj.directive.split(':', 1)
                if namespace == RECEPTOR_DIRECTIVE_NAMESPACE:
                    await directive.control(self.receptor.router, outer_env.inner_obj)
                else:
                    # other namespace/work directives
                    await self.receptor.work_manager.handle(outer_env.inner_obj)
            elif outer_env.inner_obj.message_type == 'response':
                in_response_to = outer_env.inner_obj.in_response_to
                if in_response_to in self.receptor.router.response_registry:
                    logger.info(f'Handling response to {in_response_to} with callback.')
                    for connection in self.receptor.controller_connections:
                        connection.emit_response(outer_env.inner_obj)
                else:
                    logger.warning(f'Received response to {in_response_to} but no record of sent message.')
            else:
                raise exceptions.UnknownMessageType(
                    f'Unknown message type: {outer_env.inner_obj.message_type}')
        else:
            await self.receptor.router.forward(outer_env, next_hop)


class Receptor:
    def __init__(self, config=None, node_id=None, router_cls=None,
                 work_manager_cls=None):
        self.config = config or ReceptorConfig()
        self.node_id = node_id or self.config.receptor.node_id or self._find_node_id()
        self.router = (router_cls or MeshRouter)(self)
        self.work_manager = (work_manager_cls or WorkManager)(self)
        self.connections = dict()
        self.controller_connections = []
        self.stop = False

    def _find_node_id(self):
        if 'RECEPTOR_NODE_ID' in os.environ:
            return os.environ['RECEPTOR_NODE_ID']

        node_id = uuid.uuid4()
        if os.path.exists(os.path.join(os.getcwd(), 'Pipfile')):
            with open(os.path.join(os.getcwd(), '.env'), 'w+') as ofs:
                ofs.write(f'\nRECEPTOR_NODE_ID={node_id}\n')
        return str(node_id)

    def update_connections(self, connection):
        self.router.register_edge(connection.id_, self.node_id, 1)
        if connection.id_ in self.connections:
            self.connections[connection.id_].append(connection)
        else:
            self.connections[connection.id_] = [connection]

    def add_connection(self, id_, protocol_obj):
        buffer_mgr = self.config.components.buffer_manager
        conn = Connection(id_, protocol_obj, buffer_mgr, self)
        self.update_connections(conn)
        return conn

    def remove_connection(self, conn):
        notify_protocols = []
        for connection_node in self.connections:
            if conn in self.connections[connection_node]:
                logger.info("Removing connection {} for node {}".format(conn, connection_node))
                self.connections[connection_node].remove(conn)
                self.router.update_node(self.node_id, connection_node, 100)
                self.router.debug_router()
            notify_protocols += self.connections[connection_node]
        # TODO: Broadcast update, set timer for full expiration
        for active_protocol in notify_protocols:
            active_protocol.send_route_advertisement(self.router.get_edges())

    async def shutdown_handler(self):
        while True:
            if self.stop:
                return
            await asyncio.sleep(1)



import asyncio
import logging
import socket
import sys
import os

from . import protocol

logger = logging.getLogger(__name__)


def send_directive(directive, recipient, payload, socket_path):
    if payload == '-':
        payload = sys.stdin.read()
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect(socket_path)
    sock.sendall(f"{recipient}\n{directive}\n{payload}".encode('utf-8') + protocol.DELIM)
    response = b''
    while True:
        response = sock.recv(4096)
        sys.stdout.buffer.write(response + b"\n")
        sys.stdout.flush()


def mainloop(receptor, socket_path, loop=asyncio.get_event_loop()):
    config = receptor.config
    listener = loop.create_server(
        lambda: protocol.BasicProtocol(receptor, loop),
        config.server.address, config.server.port)
    logger.info("Serving on %s:%s", config.server.address, config.server.port)
    loop.create_task(listener)
    control_listener = loop.create_unix_server(
        lambda: protocol.BasicControllerProtocol(receptor, loop),
        path=socket_path
    )
    logger.info(f'Opening control socket on {socket_path}')
    loop.create_task(control_listener)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        loop.stop()
        os.remove(socket_path)



class ReceptorRuntimeError(RuntimeError):
    pass


class ReceptorMessageError(ValueError):
    pass


class ReceptorConfigError(Exception):
    pass


class UnknownDirective(ReceptorMessageError):
    pass


class InvalidDirectiveAction(ReceptorMessageError):
    pass


class UnknownMessageType(ReceptorMessageError):
    pass


class UnrouteableError(ReceptorMessageError):
    pass
